from django.http import JsonResponse

from django_forest.utils.views.action import ActionView


class MarkAsLiveActionView(ActionView):

    def post(self, request, *args, **kwargs):
        # TODO add is_authenticated from scope pr
        params = request.GET.dict()
        body = self.get_body(request.body)
        ids = self.get_ids_from_request(request, self.Model)

        #return JsonResponse({'success': 'now live'}, safe=False)
        return JsonResponse({'html': '<h1>success</h1>'}, safe=False)



COLLECTION = {
    'name': '',
    'is_virtual': False,
    'icon': None,
    'is_read_only': False,
    'is_searchable': True,
    'only_for_relationships': False,
    'pagination_type': 'page',
    'search_fields': None,
    'actions': [],
    'segments': [],
    'fields': []
}

FIELD = {
    'field': '',
    'type': 'String',
    'is_filterable': True,
    'is_sortable': True,
    'is_read_only': False,
    'is_required': False,
    'is_virtual': False,
    'default_value': None,
    'integration': None,
    'reference': None,
    'inverse_of': None,
    'relationship': None,
    'widget': None
}

ACTION = {
    'name': '',
    'type': 'bulk',
    'baseUrl': None,
    'endpoint': '',
    'http_method': 'POST',
    'redirect': None,
    'download': False,
    'fields': [],
    'hooks': {
        'load': False,
        'change': []
    }
}

ACTION_FIELD = {
    'field': '',
    'type': '',
    'isReadOnly': False,
    'isRequired': False,
    'defaultValue': None,
    'integration': None,
    'reference': None,
    'description': None,
    'widget': None,
    'position': 0,
}



from django_forest.resources.utils.resource import ResourceView
from django_forest.utils import get_accessor_name


class AssociationView(ResourceView):
    def get_association_utils(self, Model, RelatedModel, ids):
        objects = RelatedModel.objects.filter(pk__in=ids)
        fields_to_update = [x for x in RelatedModel._meta.get_fields() if
                            x.is_relation and x.many_to_many and x.related_model == Model]

        return objects, fields_to_update

    def handle_modification(self, instance, obj, field, method):
        accessor_name = get_accessor_name(field)
        getattr(getattr(obj, accessor_name), method)(instance)

    def handle_association(self, instance, objects, fields_to_update, method):
        for field in fields_to_update:
            for obj in objects:
                self.handle_modification(instance, obj, field, method)



from django.http import JsonResponse

from django_forest.utils.views.action import ActionView


class NotExistsActionView(ActionView):
    def post(self, request, *args, **kwargs):
        body = self.get_body(request.body)
        ids = self.get_ids_from_request(request, self.Model)
        return JsonResponse({'message': 'success'})



import re
from collections import defaultdict

DICT_RE = r"^(?P<field>\w+)\[(?P<subfield>\w+)\]$"


def parse_value(value):
    if ',' in value:
        return value.split(',')
    return value

def parse_qs(parameters):
    qs = defaultdict(dict)
    for key, value in parameters.items():
        m = re.search(DICT_RE, key)
        parsed_value = parse_value(value)
        if m:
            groups = m.groupdict()
            qs[groups['field']][groups['subfield']] = parsed_value
        else:
            qs[key] = parsed_value

    return qs



from django_forest.resources.views.count import CountView
from django_forest.resources.views.csv import CsvView
from django_forest.resources.views.detail import DetailView
from django_forest.resources.views.list import ListView

__all__ = ['ListView', 'DetailView', 'CountView', 'CsvView']



from django_forest.resources.associations.utils import AssociationView
from django_forest.utils import get_association_field


class CountView(AssociationView):

    def get(self, request, pk, association_resource):
        try:
            get_association_field(self.Model, association_resource)
        except Exception as e:
            return self.error_response(e)
        else:
            queryset = getattr(self.Model.objects.get(pk=pk), association_resource).all()
            params = request.GET.dict()
            return self.get_count(queryset, params, request)



import json

from django.http import JsonResponse
from django.views import generic

from django_forest.resources.utils.queryset import QuerysetMixin
from django_forest.utils import get_association_field, get_token
from django_forest.utils.models import Models


class BaseView(QuerysetMixin, generic.View):
    def is_authenticated(self, request):
        try:
            token = get_token(request)
        except Exception:
            return False
        else:
            return token

    def get_body(self, body):
        body_unicode = body.decode('utf-8')
        return json.loads(body_unicode)

    def error_response(self, e):
        return JsonResponse({'errors': [{'detail': str(e)}]}, status=400)

    def get_model(self, resource):
        Model = Models.get(resource)
        if Model is None:
            raise Exception(f'no model found for resource {resource}')
        return Model

    def handle_all_records(self, attributes, Model, request):
        all_records_ids_excluded = attributes.get('all_records_ids_excluded', [])
        all_records_subset_query = attributes.get('all_records_subset_query', [])
        parent_collection_id = attributes.get('parent_collection_id', None)
        parent_collection_name = attributes.get('parent_collection_name', '')
        parent_association_name = attributes.get('parent_association_name', '')
        if parent_collection_id and parent_collection_name and parent_association_name:
            parent_model = Models.get(parent_collection_name)
            association_field = get_association_field(parent_model, parent_association_name)
            RelatedModel = association_field.related_model
            queryset = getattr(parent_model.objects.get(pk=parent_collection_id), parent_association_name).all()
            queryset = self.filter_queryset(queryset, RelatedModel, all_records_subset_query, request)
        else:
            queryset = Model.objects.all()
            queryset = self.filter_queryset(queryset, Model, all_records_subset_query, request)

        return [x for x in queryset.values_list('id', flat=True) if str(x) not in all_records_ids_excluded]

    def get_ids_from_request(self, request, Model):
        body = self.get_body(request.body)

        if 'data' in body:
            data = body['data']
            if 'attributes' not in data:
                return [x['id'] for x in body['data']]
            else:
                attributes = data['attributes']

                if not attributes.get('all_records', False):
                    return attributes.get('ids', [])
                else:
                    return self.handle_all_records(attributes, Model, request)
        elif 'recordIds' in body:
            return body['recordIds']



import json
import logging
import requests

from datetime import timedelta, datetime
from jose import jwt
from oic.oauth2 import AuthorizationResponse
from django.http import JsonResponse
from django.views.generic import View
from django_forest.authentication.exception import AuthenticationClientException, AuthenticationThirdPartyException

from django_forest.authentication.oidc.client_manager import OidcClientManager
from django_forest.authentication.utils import authentication_exception, get_callback_url
from django_forest.utils.error_handler import MESSAGES
from django_forest.utils.forest_api_requester import ForestApiRequester
from django_forest.utils.forest_setting import get_forest_setting


logger = logging.getLogger(__name__)


# Based on https://pyoidc.readthedocs.io/en/latest/examples/rp.html
class CallbackView(View):

    def _expiration_in_seconds(self):
        return datetime.timestamp(datetime.utcnow() + timedelta(hours=1))

    def _handle_2fa_error(self, response):
        body = response.json()
        if 'errors' in body and isinstance(body['errors'], list) and len(body['errors']) > 0:
            server_error = body['errors'][0]
            if 'name' in server_error and server_error['name'] == 'TwoFactorAuthenticationRequiredForbiddenError':
                raise AuthenticationThirdPartyException(
                    MESSAGES['SERVER_TRANSACTION']['TWO_FACTOR_AUTHENTICATION_REQUIRED']
                )

    def _handle_authent_error(self, response):
        if response.status_code == 404:
            raise AuthenticationThirdPartyException(
                MESSAGES['SERVER_TRANSACTION']['SECRET_NOT_FOUND']
            )
        elif response.status_code == 422:
            raise AuthenticationThirdPartyException(
                MESSAGES['SERVER_TRANSACTION']['SECRET_AND_RENDERINGID_INCONSISTENT']
            )
        else:
            body = response.json()
            if isinstance(body.get('errors'), list) and len(body['errors']) > 0:
                server_error = body['errors'][0]
                if server_error.get('name') == 'TwoFactorAuthenticationRequiredForbiddenError':
                    raise AuthenticationThirdPartyException(
                        MESSAGES['SERVER_TRANSACTION']['TWO_FACTOR_AUTHENTICATION_REQUIRED']
                    )

        logger.warning(f"Unknown error on the authentication process {response.status_code} {response.json()}")
        raise AuthenticationThirdPartyException(
            MESSAGES['SERVER_TRANSACTION']['AUTHORIZATION']
        )

    def parse_authorization_response(self, client, state, full_path_info):
        return client.parse_response(
            AuthorizationResponse,
            info=full_path_info,
            sformat='urlencoded',
            state=state,
            scope=['openid', 'email', 'profile']
        )

    def _get_state_params(self, request):
        try:
            state = json.loads(
                request.GET['state']
            )
        except KeyError:
            raise AuthenticationClientException(
                MESSAGES['SERVER_TRANSACTION']['INVALID_STATE_MISSING']
            )
        except ValueError:
            raise AuthenticationClientException(
                MESSAGES['SERVER_TRANSACTION']['INVALID_STATE_FORMAT']
            )
        return state

    def _authenticate(self, client, state, code):
        access_token_request = client.do_access_token_request(
            state=json.dumps(state),
            scope=['openid', 'email', 'profile'],
            authn_method='',
            request_args={'code': code},
            verify=False
        )
        rendering_id = state['renderingId']
        route = f'/liana/v2/renderings/{rendering_id}/authorization'

        url = ForestApiRequester.build_url(route)
        headers = {'forest-token': access_token_request['access_token']}

        try:
            response = ForestApiRequester.get(
                url,
                headers=headers
            )
        except requests.exceptions.RequestException:
            raise AuthenticationThirdPartyException(
                MESSAGES['SERVER_TRANSACTION']['AUTHORIZATION']
            )
        if response.status_code == 200:
            body = response.json()
            user = body['data']['attributes']
            user['id'] = body['data']['id']
            return user
        else:
            self._handle_authent_error(response)

    def _verify_code_and_generate_token_body(self, redirect_url, request):
        client = OidcClientManager.get_client_for_callback_url(redirect_url)
        client.do_access_token_request
        state = self._get_state_params(request)
        if 'renderingId' not in state:
            raise AuthenticationClientException(MESSAGES['SERVER_TRANSACTION']['INVALID_STATE_RENDERING_ID'])
        authorization_resp = self.parse_authorization_response(
            client,
            state,
            request.get_full_path_info(),
        )
        user = self._authenticate(
            client,
            state,
            authorization_resp['code']
        )
        return {
            'id': user['id'],
            'email': user['email'],
            'first_name': user['first_name'],
            'last_name': user['last_name'],
            'team': user['teams'][0],
            'rendering_id': state['renderingId'],
            'exp': self._expiration_in_seconds()
        }

    @authentication_exception
    def get(self, request, *args, **kwargs):
        callback_url = get_callback_url()
        token_body = self._verify_code_and_generate_token_body(callback_url, request)
        auth_secret = get_forest_setting('FOREST_AUTH_SECRET')
        return JsonResponse({
            'token': jwt.encode(
                token_body,
                auth_secret,
                algorithm='HS256'
            ),
            'tokenData': token_body
        })



import copy
from django.test import TestCase

from django_forest.tests.fixtures.schema import test_schema

from django_forest.utils.collection import Collection
from django_forest.tests.models import Question
from django_forest.utils.schema import Schema
from django_forest.utils.scope import ScopeManager


class QuestionForest(Collection):
    def load(self):
        self.fields = [
            {
                'field': 'foo',
                'type': 'String',
                'get': self.get_foo,
                'set': self.set_foo,
            },
            {
                'field': 'question_text',  # override existing field
                'type': 'String',
            }
        ]
        self.segments = [
            {
                'name': 'Best questions',
                'where': self.best_questions
            }
        ]

        self.actions = [
            {
                'name': 'Mark as Live',
            },
            {
                'name': 'Upload Legal Docs',
                'endpoint': '/foo/upload_legal_docs',
                'http_method': 'GET',
                'download': True,
                'type': 'single',
                'fields': [
                    {
                        'field': 'Certificate of Incorporation',
                        'description': 'The legal document relating to the formation of a company or corporation.',
                        'type': 'File',
                        'isRequired': True
                    }, {
                        'field': 'Proof of address',
                        'description': '(Electricity, Gas, Water, Internet, Landline & Mobile Phone Invoice / Payment Schedule) no older than 3 months of the legal representative of your company',
                        'type': 'File',
                        'isRequired': True
                    }, {
                        'field': 'Company bank statement',
                        'description': 'PDF including company name as well as IBAN',
                        'type': 'File',
                        'isRequired': True
                    }, {
                        'field': 'Valid proof of ID',
                        'description': 'ID card or passport if the document has been issued in the EU, EFTA, or EEA / ID card or passport + resident permit or driving licence if the document has been issued outside the EU, EFTA, or EEA of the legal representative of your company',
                        'type': 'File',
                        'isRequired': True
                    }
                ]
            },
            {
                'name': 'Send invoice',
                'type': 'single',
                'fields': [
                    {
                        'field': 'country',
                        'type': 'Enum',
                        'enums': ['FR', 'US']
                    },
                    {
                        'field': 'city',
                        'type': 'String',
                        'hook': 'cityChange'
                    },
                    {
                        'field': 'zip code',
                        'type': 'String',
                        'hook': 'zipCodeChange'
                    },
                ],
                'hooks': {
                    'load': self.invoice_load,
                    'change': {
                        'cityChange': self.invoice_change_city,
                        'zipCodeChange': self.invoice_change_zip_code,
                    },
                },
            }
        ]

    def get_foo(self, obj):
        return obj.question_text + 'foo'

    def set_foo(self, obj, value):
        obj.question_text = f'{value}-foo'
        return obj

    def best_questions(self, obj):
        questions = Question.objects.raw('''SELECT tests_question.id, COUNT(tests_choices.*)
            FROM tests_question
            JOIN tests_choices ON tests_choices.question_id = question.id
            GROUP BY question.id
            ORDER BY count DESC
            LIMIT 5;''')
        return {'id__in': [question.id for question in questions]}

    def invoice_load(self, fields, request, *args, **kwargs):
        country = next((x for x in fields if x['field'] == 'country'), None)
        country['value'] = 'FR'
        return fields

    def invoice_change_city(self, fields, request, changed_field, *args, **kwargs):
        zip_code = next((x for x in fields if x['field'] == 'zip code'), None)
        zip_code['value'] = '83'
        return fields

    def invoice_change_zip_code(self, fields, request, changed_field, *args, **kwargs):
        return fields


class CustomNameForest(Collection):
    name = 'custom'


class UtilsCollectionTests(TestCase):

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Collection._registry = {}

    def tearDown(self):
        # reset _registry after each test
        Collection._registry = {}
        ScopeManager.cache = {}

    def test_register(self):
        Collection.register(QuestionForest, Question)
        self.assertEqual(len(Collection._registry), 1)
        self.assertIsInstance(Collection._registry['tests_question'], Collection)

    def test_register_smart_collection(self):
        Collection.register(QuestionForest)
        self.assertEqual(len(Collection._registry), 1)
        self.assertIsInstance(Collection._registry['QuestionForest'], Collection)
        collection = Schema.get_collection('QuestionForest')
        self.assertTrue(collection['is_virtual'])

    def test_register_smart_collection_custom_name(self):
        Collection.register(CustomNameForest)
        self.assertEqual(len(Collection._registry), 1)
        self.assertIsInstance(Collection._registry['custom'], Collection)

    def test_register_smart_field(self):
        Collection.register(QuestionForest, Question)
        collection = Schema.get_collection('tests_question')
        self.assertEqual(len(collection['fields']), 6)
        foo_smart_field = [f for f in collection['fields'] if f['field'] == 'foo'][0]
        self.assertTrue(foo_smart_field['is_virtual'])

    def test_register_smart_field_override(self):
        Collection.register(QuestionForest, Question)
        self.assertEqual(len(Collection._registry), 1)
        self.assertIsInstance(Collection._registry['tests_question'], Collection)
        collection = Schema.get_collection('tests_question')
        question_text_field = [f for f in collection['fields'] if f['field'] == 'question_text'][0]
        self.assertTrue(question_text_field['is_virtual'])

    def test_register_smart_segment(self):
        Collection.register(QuestionForest, Question)
        collection = Schema.get_collection('tests_question')
        self.assertEqual(len(collection['segments']), 1)
        self.assertEqual(collection['segments'][0], {'name': 'Best questions'})

    def test_register_smart_action(self):
        Collection.register(QuestionForest, Question)
        collection = Schema.get_collection('tests_question')
        self.assertEqual(len(collection['actions']), 3)
        self.assertEqual(collection['actions'][0], {
            'name': 'Mark as Live',
            'type': 'bulk',
            'baseUrl': None,
            'endpoint': '/forest/actions/mark-as-live',
            'http_method': 'POST',
            'redirect': None,
            'download': False,
            'fields': [],
            'hooks': {
                'load': False,
                'change': []
            }
        })
        self.assertEqual(collection['actions'][1], {
            'name': 'Upload Legal Docs',
            'type': 'single',
            'baseUrl': None,
            'endpoint': '/foo/upload_legal_docs',
            'http_method': 'GET',
            'redirect': None,
            'download': True,
            'fields': [
                {
                    'field': 'Certificate of Incorporation',
                    'description': 'The legal document relating to the formation of a company or corporation.',
                    'type': 'File',
                    'isRequired': True,
                    'isReadOnly': False,
                    'defaultValue': None,
                    'integration': None,
                    'reference': None,
                    'widget': None,
                    'position': 0
                },
                {
                    'field': 'Proof of address',
                    'description': '(Electricity, Gas, Water, Internet, Landline & Mobile Phone Invoice / Payment Schedule) no older than 3 months of the legal representative of your company',
                    'type': 'File',
                    'isRequired': True,
                    'isReadOnly': False,
                    'defaultValue': None,
                    'integration': None,
                    'reference': None,
                    'widget': None,
                    'position': 1
                },
                {
                    'field': 'Company bank statement',
                    'description': 'PDF including company name as well as IBAN',
                    'type': 'File',
                    'isRequired': True,
                    'isReadOnly': False,
                    'defaultValue': None,
                    'integration': None,
                    'reference': None,
                    'widget': None,
                    'position': 2
                },
                {
                    'field': 'Valid proof of ID',
                    'description': 'ID card or passport if the document has been issued in the EU, EFTA, or EEA / ID card or passport + resident permit or driving licence if the document has been issued outside the EU, EFTA, or EEA of the legal representative of your company',
                    'type': 'File',
                    'isRequired': True,
                    'isReadOnly': False,
                    'defaultValue': None,
                    'integration': None,
                    'reference': None,
                    'widget': None,
                    'position': 3
                }],
            'hooks': {
                'load': False,
                'change': []
            }
        })
        self.assertEqual(collection['actions'][2], {
            'name': 'Send invoice',
            'type': 'single',
            'baseUrl': None,
            'endpoint': '/forest/actions/send-invoice',
            'http_method': 'POST',
            'redirect': None,
            'download': False,
            'fields': [
                {
                    'field': 'country',
                    'type': 'Enum',
                    'enums': ['FR', 'US'],
                    'isReadOnly': False,
                    'isRequired': False,
                    'defaultValue': None,
                    'integration': None,
                    'reference': None,
                    'description': None,
                    'widget': None,
                    'position': 0
                },
                {
                    'field': 'city',
                    'type': 'String',
                    'isReadOnly': False,
                    'isRequired': False,
                    'defaultValue': None,
                    'integration': None,
                    'reference': None,
                    'description': None,
                    'widget': None,
                    'position': 1,
                    'hook': 'cityChange'
                }, {
                    'field': 'zip code',
                    'type': 'String',
                    'isReadOnly': False,
                    'isRequired': False,
                    'defaultValue': None,
                    'integration': None,
                    'reference': None,
                    'description': None,
                    'widget': None,
                    'position': 2,
                    'hook': 'zipCodeChange'
                }],
            'hooks': {
                'load': True,
                'change': ['cityChange', 'zipCodeChange']
            }
        })



from django.http import HttpResponse

from django_forest.utils import get_token
from django_forest.utils.permissions import Permission
from django_forest.utils.views.base import BaseView


class ActionView(BaseView):
    def get_smart_action_request_info(self, request):
        return {
            'endpoint': request.path,
            'http_method': request.method
        }

    def is_authorized(self, request, token, resource):
        smart_action_request_info = self.get_smart_action_request_info(request)
        permission = Permission(resource,
                                'actions',
                                token['rendering_id'],
                                token['id'],
                                smart_action_request_info=smart_action_request_info)
        if not Permission.is_authorized(permission):
            raise Exception('not authorized')

    def dispatch(self, request, *args, **kwargs):
        body = self.get_body(request.body)
        resource = body['data']['attributes']['collection_name']

        try:
            self.Model = self.get_model(resource)
        except Exception as e:
            return self.error_response(e)
        else:
            # check permissions
            try:
                token = get_token(request)
                self.is_authorized(request, token, resource)
            except Exception:
                return HttpResponse(status=403)
            else:
                return super().dispatch(request, *args, **kwargs)



from django_forest.tests.utils.test_forest_api_requester import mocked_requests

mocked_ip_whitelist = {
    'data': {
        'type': 'ip-whitelist-rules',
        'id': '1',
        'attributes': {
            'rules': [],
            'use_ip_whitelist': False
        }
    }
}

mocked_permissions = {
    'data': {
        'collections': {
            'tests_question': {
                'collection': {
                    'browseEnabled': True,
                    'readEnabled': True,
                    'addEnabled': True,
                    'editEnabled': True,
                    'deleteEnabled': True,
                    'exportEnabled': True
                },
                'actions': {}
            },
        },
        'renderings': {
            '1': {}
        }
    },
    'stats': {
        'queries': [],
        'leaderboards': [],
        'lines': [],
        'objectives': [],
        'percentages': [],
        'pies': [],
        'values': []
    },
    'meta': {
        'rolesACLActivated': True
    }
}

mocked_config = {
    'permissions': {
        'data': mocked_permissions,
        'status': 200
    },
    'ip_whitelist': {
        'data': mocked_ip_whitelist,
        'status': 200
    },
    'scope': {
        'data': {},
        'status': 200
    }
}


def mocked_requests_permission(value, *args):
    def m(url, **kwargs):
        if url == 'https://api.test.forestadmin.com/liana/v3/permissions':
            return mocked_requests(value['permissions']['data'], value['permissions']['status'])
        elif url == 'https://api.test.forestadmin.com/liana/scopes':
            return mocked_requests(value['scope']['data'], value['scope']['status'])
        elif url == 'https://api.test.forestadmin.com/liana/v1/ip-whitelist-rules':
            return mocked_requests(value['ip_whitelist']['data'], value['ip_whitelist']['status'])

    return m



from typing import Callable, Optional

from datetime import datetime, timedelta, date

from django_forest.resources.utils.queryset.filters.date.utils import (
    get_previous_x_days,
    get_previous_x_weeks,
    get_previous_x_months,
    get_previous_x_quarters,
    get_previous_x_years,
)
from django_forest.resources.utils.queryset.filters.date.conditions import (
    RangeCondition,
    LowerThanCondition,
    LowerThanOrEqualCondition,
    GreaterThanCondition,
    GreaterThanOrEqualCondition,
)


class ConditionFactory:

    DATE_CONDITION_BUILDERS = {
        'today': 'build_today_condition',
        'yesterday': 'build_yesterday_condition',
        'previous_x_days': 'build_previous_x_days_condition',
        'previous_x_days_to_date': 'build_previous_x_days_to_date_condition',
        'previous_week': 'build_previous_week_condition',
        'previous_week_to_date': 'build_previous_week_to_date_condition',
        'previous_month': 'build_previous_month_condition',
        'previous_month_to_date': 'build_previous_month_to_date_condition',
        'previous_quarter': 'build_previous_quarter_condition',
        'previous_quarter_to_date': 'build_previous_quarter_to_date_condition',
        'previous_year': 'build_previous_year_condition',
        'previous_year_to_date': 'build_previous_year_to_date_condition',
        # Don't handle offset
        'before_x_hours_ago': 'build_before_x_hours_ago_condition',
        'after_x_hours_ago': 'build_after_x_hours_ago_condition',
        'past': 'build_past_condition',
        'future': 'build_future_condition',
    }

    OPERATORS = list(DATE_CONDITION_BUILDERS.keys())
    OFFSET_OPERATORS = OPERATORS[:-4]

    def __init__(self, dt: datetime, *args, **kwargs) -> None:
        self.datetime = dt

    def _get_tomorrow(self, current_datetime: datetime) -> date:
        return current_datetime + timedelta(days=1)

    def _handle_to_date(self, current_datetime: datetime, range: RangeCondition) -> RangeCondition:
        start = range.end
        end = self._get_tomorrow(current_datetime)
        return RangeCondition(start, end)

    def build_today_condition(self, current_datetime: datetime, offset: int, **kwargs) -> RangeCondition:
        current_datetime = self._get_tomorrow(current_datetime) + timedelta(days=-offset)
        return RangeCondition(*get_previous_x_days(current_datetime, 1))

    def build_yesterday_condition(self, current_datetime: datetime, offset: int, **kwargs) -> RangeCondition:
        return RangeCondition(*get_previous_x_days(current_datetime, 1, offset))

    def build_previous_x_days_condition(self, current_datetime: datetime, offset: int, period: int) -> RangeCondition:
        return RangeCondition(*get_previous_x_days(current_datetime, period, offset))

    def build_previous_x_days_to_date_condition(
        self,
        current_datetime: datetime,
        offset: int,
        period: int
    ) -> RangeCondition:
        return self.build_previous_x_days_condition(
            self._get_tomorrow(current_datetime),
            offset,
            period,
        )

    def build_previous_week_condition(
        self,
        current_datetime: datetime,
        offset: int,
        period: int = 1
    ) -> RangeCondition:
        return RangeCondition(*get_previous_x_weeks(current_datetime, period, offset))

    def build_previous_week_to_date_condition(
        self,
        current_datetime: datetime,
        offset: int,
        **kwargs,
    ) -> RangeCondition:
        if not offset:
            return self._handle_to_date(
                current_datetime,
                self.build_previous_week_condition(
                    current_datetime,
                    offset
                )
            )
        return self.build_previous_week_condition(
            current_datetime,
            offset - 1
        )

    def build_previous_month_condition(
        self,
        current_datetime: datetime,
        offset: int,
        period: int = 1
    ) -> RangeCondition:
        return RangeCondition(*get_previous_x_months(current_datetime, period, offset))

    def build_previous_month_to_date_condition(
        self,
        current_datetime: datetime,
        offset: int,
        **kwargs
    ) -> RangeCondition:
        if not offset:
            return self._handle_to_date(
                current_datetime,
                self.build_previous_month_condition(current_datetime, offset)
            )
        return self.build_previous_month_condition(
            current_datetime,
            offset - 1
        )

    def build_previous_quarter_condition(
        self,
        current_datetime: datetime,
        offset: int,
        period: int = 1,
    ) -> RangeCondition:
        return RangeCondition(*get_previous_x_quarters(current_datetime, period, offset))

    def build_previous_quarter_to_date_condition(
        self,
        current_datetime: datetime,
        offset: int,
        **kwargs
    ) -> RangeCondition:
        if not offset:
            return self._handle_to_date(
                current_datetime,
                self.build_previous_quarter_condition(current_datetime, offset)
            )
        return self.build_previous_quarter_condition(
            current_datetime,
            offset - 1
        )

    def build_previous_year_condition(self, current_datetime: datetime, offset: int, period: int = 1) -> RangeCondition:
        return RangeCondition(*get_previous_x_years(current_datetime, period, offset))

    def build_previous_year_to_date_condition(
        self,
        current_datetime: datetime,
        offset: int,
        **kwargs
    ) -> RangeCondition:
        if not offset:
            return self._handle_to_date(
                current_datetime,
                self.build_previous_year_condition(current_datetime, offset)
            )
        return self.build_previous_year_condition(
            current_datetime,
            offset - 1
        )

    def build_before_x_hours_ago_condition(
        self,
        current_datetime: datetime,
        period: int,
        **kwargs
    ) -> LowerThanCondition:
        return LowerThanCondition(current_datetime + timedelta(hours=-period))

    def build_after_x_hours_ago_condition(
        self,
        current_datetime: datetime,
        period: int,
        **kwargs
    ) -> GreaterThanCondition:
        return GreaterThanCondition(
            current_datetime + timedelta(hours=period)
        )

    def build_past_condition(self, current_datetime: datetime, **kwargs) -> LowerThanOrEqualCondition:
        return LowerThanOrEqualCondition(
            current_datetime
        )

    def build_future_condition(self, current_datetime: datetime, **kwargs) -> GreaterThanOrEqualCondition:
        return GreaterThanOrEqualCondition(
            current_datetime
        )

    def get_condition_builder(self, condition_name: str) -> Callable:
        method_name = self.DATE_CONDITION_BUILDERS.get(condition_name)
        res = None
        if method_name:
            res = getattr(self, method_name)
        return res

    def build(self, condition_name: str, period: Optional[int] = None, offset: Optional[int] = None) -> RangeCondition:
        date_condition = None
        build_condition = self.get_condition_builder(condition_name)
        if build_condition:
            kwargs = {
                'current_datetime': self.datetime,
            }
            if period:
                kwargs['period'] = period
            if offset is not None:
                kwargs['offset'] = offset
            date_condition = build_condition(**kwargs)
        return date_condition



from django.urls import path, include
from django.views.decorators.csrf import csrf_exempt

from . import views
from .associations import urls as associations_urls

app_name = 'resources'
urlpatterns = [
    path('', csrf_exempt(views.ListView.as_view()), name='list'),
    path('.csv', csrf_exempt(views.CsvView.as_view()), name='csv'),
    path('/count', views.CountView.as_view(), name='count'),
    path('/<pk>', csrf_exempt(views.DetailView.as_view()), name='detail'),
    path('/<pk>/relationships', include(associations_urls)),
]



from django_forest.utils.cors import set_cors
from django_forest.utils.middlewares import set_middlewares
from django_forest.utils.schema import Schema


def init_forest():
    set_cors()
    set_middlewares()

    # schema
    Schema.build_schema()
    Schema.add_smart_features()
    Schema.handle_json_api_schema()
    Schema.handle_schema_file()
    Schema.send_apimap()



import logging

# 5 minutes expiration cache
from django_forest.utils.date import get_utc_now

from django_forest.utils.forest_api_requester import ForestApiRequester
from django_forest.utils.permissions import date_difference_in_seconds

SCOPE_CACHE_EXPIRATION_DELTA = 60 * 5

# Get an instance of a logger
logger = logging.getLogger(__name__)


class ScopeManager:
    cache = {}

    @classmethod
    def _has_cache_expired(cls, rendering_id):
        if rendering_id not in cls.cache:
            return True
        rendering_scopes = cls.cache[rendering_id]
        seconds_since_last_fetch = date_difference_in_seconds(get_utc_now(), rendering_scopes['fetched_at'])
        return seconds_since_last_fetch > SCOPE_CACHE_EXPIRATION_DELTA

    @classmethod
    def _refresh_cache(cls, rendering_id):
        try:
            scopes = ForestApiRequester.get_from_rendering_id('/liana/scopes', rendering_id)
        except Exception:
            raise Exception('Unable to fetch scopes')
        else:
            cls.cache[rendering_id] = {
                'scopes': scopes,
                'fetched_at': get_utc_now()
            }

    @staticmethod
    def _format_dynamic_values(user_id, collection_scope):
        try:
            scope = collection_scope['scope']
            for condition in scope['filter']['conditions']:
                if condition['value'].startswith('$currentUser'):
                    condition['value'] = scope['dynamicScopesValues']['users'][user_id][condition['value']]
            return scope['filter']
        except Exception:
            return None

    @classmethod
    def _get_scope_collection_scope(cls, rendering_id, collection_name):
        # TODO: handle cache stale true, do not wait for requests if cache expired using a ThreadPoolExecutor
        # https://stackoverflow.com/questions/14245989/python-requests-non-blocking
        if cls._has_cache_expired(rendering_id):
            cls._refresh_cache(rendering_id)

        if collection_name in cls.cache[rendering_id]['scopes']:
            return cls.cache[rendering_id]['scopes'][collection_name]

    @classmethod
    def get_scope_for_user(cls, token, collection_name):
        if 'rendering_id' not in token:
            raise Exception('Missing required rendering_id')

        collection_scope = cls._get_scope_collection_scope(str(token['rendering_id']), collection_name)
        return cls._format_dynamic_values(str(token['id']), collection_scope)

    @classmethod
    def invalidate_scope_cache(cls, rendering_id):
        if rendering_id in cls.cache:
            del cls.cache[rendering_id]



from django_forest.actions.hooks.views.utils import HookView


class LoadView(HookView):
    def handle_action(self, action, action_name, request):
        if self.hook_exists(action, action_name, 'load'):
            return action['hooks']['load'](action['fields'], request)



from datetime import datetime

import pytz

from django_forest.utils.forest_api_requester import ForestApiRequester
from django_forest.utils.forest_setting import get_forest_setting
from django_forest.utils.permissions.utils import date_difference_in_seconds, is_stat_allowed, is_user_allowed,\
    is_smart_action_allowed


class Permission:
    renderings_cached = {}
    permissions_cached = {}
    expiration_in_seconds = get_forest_setting('FOREST_PERMISSIONS_EXPIRATION_IN_SECONDS', 3600)

    def __init__(self, *args, **kwargs):
        self.collection_name = args[0]
        self.permission_name = args[1]
        self.rendering_id = str(args[2])
        self.user_id = str(args[3])

        self.query_request_info = kwargs.get('query_request_info', None)
        self.smart_action_request_info = kwargs.get('smart_action_request_info', None)

    @classmethod
    def is_authorized(cls, obj):
        if not cls.have_permissions_expired() and cls.is_allowed(obj):
            return True

        # Notice fetch if permissions have expired or not allowed, to get last update
        cls.fetch_permissions(obj.rendering_id)

        return cls.is_allowed(obj)

    @classmethod
    def have_permissions_expired(cls):
        if 'last_fetch' in cls.permissions_cached:
            elapsed_seconds = date_difference_in_seconds(datetime.now(pytz.UTC), cls.permissions_cached['last_fetch'])
            return elapsed_seconds >= cls.expiration_in_seconds

        return True

    @classmethod
    def fetch_permissions(cls, rendering_id):
        permissions = ForestApiRequester.get_from_rendering_id('/liana/v3/permissions', rendering_id)
        permissions['last_fetch'] = datetime.now(pytz.UTC)
        cls.permissions_cached = permissions

    @classmethod
    def is_allowed(cls, obj):
        # NOTICE: check liveQueries permissions
        if obj.permission_name == 'liveQueries':
            return cls.live_query_allowed(obj)
        elif obj.permission_name == 'statsWithParameters':
            return cls.stat_with_parameters_allowed(obj)

        return cls.permission_allowed(obj)

    @classmethod
    def permission_allowed(cls, obj):
        try:
            permissions = cls.permissions_cached['data']['collections'][obj.collection_name]
        except Exception:
            return False
        else:
            if obj.permission_name == 'actions':
                return is_smart_action_allowed(obj, permissions['actions'])
            else:
                return is_user_allowed(obj.user_id, permissions['collection'][obj.permission_name])

    @classmethod
    def live_query_allowed(cls, obj):
        permissions = cls.permissions_cached
        try:
            live_queries_permissions = permissions['stats']['queries']
        except Exception:
            return False
        else:
            # NOTICE: query_request_info matching an existing live query
            return obj.query_request_info in live_queries_permissions

    @classmethod
    def stat_with_parameters_allowed(cls, obj):
        permission_type = f"{obj.query_request_info['type'].lower()}s"
        permissions = cls.permissions_cached

        try:
            pool_permissions = permissions['stats'][permission_type]
        except Exception:
            return False
        else:
            array_permission_infos = [x for x in obj.query_request_info.values() if x is not None]
            return is_stat_allowed(pool_permissions, array_permission_infos)



from django.urls import path, include

from .. import urls as django_forest_urls
from .actions import urls as actions_urls

app_name = 'tests'
urlpatterns = [
    path('forest/actions', include(actions_urls)),  # /!\ actions must be first
    path('forest', include(django_forest_urls)),
]



from django.conf import settings


def set_middlewares():
    settings.MIDDLEWARE.insert(0, 'django_forest.middleware.PermissionMiddleware')
    settings.MIDDLEWARE.insert(0, 'django_forest.middleware.IpWhitelistMiddleware')



from distutils.util import strtobool

from django.http import JsonResponse, HttpResponse

from django_forest.resources.associations.utils import AssociationView
from django_forest.resources.utils.json_api_serializer import JsonApiSerializerMixin
from django_forest.resources.utils.smart_field import SmartFieldMixin
from django_forest.utils import get_association_field


class ListView(SmartFieldMixin, JsonApiSerializerMixin, AssociationView):

    def get(self, request, pk, association_resource):
        try:
            association_field = get_association_field(self.Model, association_resource)
        except Exception as e:
            return self.error_response(e)
        else:
            RelatedModel = association_field.related_model

            # default
            queryset = getattr(self.Model.objects.get(pk=pk), association_resource).all()

            params = request.GET.dict()
            # enhance queryset
            queryset = self.enhance_queryset(queryset, RelatedModel, params, request)

            # handle smart fields
            self.handle_smart_fields(queryset, RelatedModel._meta.db_table, many=True)

            # json api serializer
            data = self.serialize(queryset, RelatedModel, params)

            return JsonResponse(data, safe=False)

    def post(self, request, pk, association_resource):
        try:
            association_field = get_association_field(self.Model, association_resource)
        except Exception as e:
            return self.error_response(e)
        else:
            RelatedModel = association_field.related_model
            body = self.get_body(request.body)
            ids = [x['id'] for x in body['data']]

            instance = self.Model.objects.get(pk=pk)
            objects, fields_to_update = self.get_association_utils(self.Model, RelatedModel, ids)
            self.handle_association(instance, objects, fields_to_update, 'add')
            return JsonResponse({}, safe=False)

    # Notice: BelongsTo case
    def put(self, request, pk, association_resource):
        # Notice, ForeignKeys are already handled by the put method in the resource view
        # There is no need to do anything there
        return HttpResponse(status=204)

    def _dissociate(self, RelatedModel, ids, pk):
        instance = self.Model.objects.get(pk=pk)
        objects, fields_to_update = self.get_association_utils(self.Model, RelatedModel, ids)
        self.handle_association(instance, objects, fields_to_update, 'remove')

    def delete(self, request, pk, association_resource):
        try:
            association_field = get_association_field(self.Model, association_resource)
            ids = self.get_ids_from_request(request, self.Model)
        except Exception as e:
            return self.error_response(e)
        else:
            RelatedModel = association_field.related_model
            params = request.GET.dict()

            # Notice: Delete
            if 'delete' in params and strtobool(params['delete']):
                # Notice: this does not run pre/post_delete signals
                RelatedModel.objects.filter(pk__in=ids).delete()
            # Notice: Dissociate
            else:
                self._dissociate(RelatedModel, ids, pk)

            return HttpResponse(status=204)



# flake8: noqa

MESSAGES = {
    'CONFIGURATION': {
        'AUTH_SECRET_MISSING': 'Your Forest authSecret seems to be missing. Can you check that you properly set a Forest authSecret in the Forest initializer?',
    },
    'SERVER_TRANSACTION': {
        'SECRET_AND_RENDERINGID_INCONSISTENT': 'Cannot retrieve the project you\'re trying to unlock. The envSecret and renderingId seems to be missing or inconsistent.',
        'SERVER_DOWN': 'Cannot retrieve the data from the Forest server. Forest API seems to be down right now.',
        'SECRET_NOT_FOUND': 'Cannot retrieve the data from the Forest server. Can you check that you properly copied the Forest envSecret in your settings?',
        'UNEXPECTED': 'Cannot retrieve the data from the Forest server. An error occurred in Forest API.',
        'INVALID_STATE_MISSING': 'Invalid response from the authentication server: the state parameter is missing',
        'INVALID_STATE_FORMAT': 'Invalid response from the authentication server: the state parameter is not at the right format',
        'INVALID_STATE_RENDERING_ID': 'Invalid response from the authentication server: the state does not contain a renderingId',
        'MISSING_RENDERING_ID': 'Authentication request must contain a renderingId',
        'INVALID_RENDERING_ID': 'The parameter renderingId is not valid',
        'REGISTRATION_FAILED': 'The registration to the authentication API failed, response: ',
        'OIDC_CONFIGURATION_RETRIEVAL_FAILED': 'Failed to retrieve the provider\'s configuration.',
        'TWO_FACTOR_AUTHENTICATION_REQUIRED': 'Two factor authentication required',
        'AUTHORIZATION': 'Error while authorizing the user on Forest Admin'
    }
}



from urllib.parse import urljoin
from django.urls import reverse

from django.http import JsonResponse
from django_forest.authentication.exception import AuthenticationSettingsException, BaseAuthenticationException
from django_forest.utils.forest_setting import get_forest_setting


def get_callback_url():
    application_url = get_forest_setting('APPLICATION_URL')

    if not application_url:
        raise AuthenticationSettingsException('APPLICATION_URL is not defined in your FOREST settings')

    return urljoin(application_url, reverse('django_forest:authentication:callback'))


def authentication_exception(f):
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except BaseAuthenticationException as error:
            return JsonResponse({'errors': [{'detail': str(error)}]}, status=error.STATUS)
    return wrapper



import json
import sys
from unittest import mock

import pytest
from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema


# reset forest config dir auto import
@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


@pytest.mark.usefixtures('reset_config_dir_import')
class LoadHooksViewTests(TransactionTestCase):
    fixtures = ['question.json']

    def setUp(self):
        Schema.build_schema()
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:actions:hooks:load', kwargs={'action_name': 'send-invoice'})
        self.not_exists_url = reverse('django_forest:actions:hooks:load', kwargs={'action_name': 'not-exists'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_load(self, mocked_decode):
        body = {
            'recordIds': ['70'],
            'collectionName': 'tests_question'
        }
        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(data, {
            'fields': [
                {'field': 'country', 'type': 'Enum', 'enums': ['FR', 'US'], 'isReadOnly': False, 'isRequired': False,
                 'defaultValue': None, 'integration': None, 'reference': None, 'description': None, 'widget': None,
                 'position': 0, 'value': None},
                {'field': 'phones', 'type': ['Enum'], 'enums': [['01', '02'], ['998', '999']], 'isReadOnly': False,
                 'isRequired': False, 'defaultValue': None, 'integration': None, 'reference': None, 'description': None,
                 'widget': None, 'position': 1, 'value': None},
                {'field': 'city', 'type': 'String', 'hook': 'cityChange', 'isReadOnly': False, 'isRequired': False,
                 'defaultValue': None, 'integration': None, 'reference': None, 'description': None, 'widget': None,
                 'position': 2}, {'field': 'zip code', 'type': 'String', 'hook': 'zipCodeChange', 'isReadOnly': False,
                                  'isRequired': False, 'defaultValue': None, 'integration': None, 'reference': None,
                                  'description': None, 'widget': None, 'position': 3}]})

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    def test_load_not_exists(self, mocked_decode):
        body = {
            'recordIds': ['70'],
            'collectionName': 'tests_question'
        }
        response = self.client.post(self.not_exists_url, json.dumps(body), content_type='application/json')
        self.assertEqual(response.status_code, 400)
        data = response.json()
        self.assertEqual(data, {'errors': [{'detail': 'action not found'}]})


@pytest.mark.usefixtures('reset_config_dir_import')
class LoadHooksViewNoTokenTests(TransactionTestCase):
    fixtures = ['question.json']

    def setUp(self):
        Schema.build_schema()
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:actions:hooks:load', kwargs={'action_name': 'send-invoice'})

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}

    def test_load(self):
        body = {
            'recordIds': ['70'],
            'collectionName': 'tests_question'
        }
        response = self.client.post(self.url, json.dumps(body), content_type='application/json')
        self.assertEqual(response.status_code, 403)



import json
import logging
import requests

from django_forest.authentication.exception import AuthenticationThirdPartyException

from django_forest.utils.error_handler import MESSAGES
from django_forest.utils.forest_api_requester import ForestApiRequester
from django_forest.utils.forest_setting import get_forest_setting

logger = logging.getLogger(__name__)

def handle_register_error(response):
    data = response.json()
    logger.warning(
        f"Oic client registration error {data}"
    )
    if 'error' in data:
        raise AuthenticationThirdPartyException(data)
    raise AuthenticationThirdPartyException(
        MESSAGES['SERVER_TRANSACTION']['REGISTRATION_FAILED'] + json.dumps(data)
    )

def register(metadata):
    initial_access_token = get_forest_setting('FOREST_ENV_SECRET')
    headers = {}
    if initial_access_token:
        headers = {'Authorization': f'Bearer {initial_access_token}'}
    try:
        response = ForestApiRequester.post(metadata['registration_endpoint'], body=metadata, headers=headers)
    except requests.exceptions.RequestException:
        raise AuthenticationThirdPartyException(
            MESSAGES['SERVER_TRANSACTION']['REGISTRATION_FAILED']
        )
    else:
        if response.status_code != 201:
            handle_register_error(response)
        else:
            logger.info(
                f"New oic client registration success. \n metadata: {metadata}, \n headers: {headers}"
            )

    return response.json()



from .in_search_fields import in_search_fields
from django_forest.utils.schema import Schema


class DecoratorsMixin:
    def get_fields_for_decorator_search(self, collection):
        fields_to_search = []
        for x in collection['fields']:
            if x['type'] in ('String', 'Number', 'Enum') \
                    and not x['reference'] \
                    and in_search_fields(x['field'], collection['search_fields']):
                fields_to_search.append(x)
        return fields_to_search

    def handle_search_decorator_field(self, field, record, data, search):
        if field['field'] in record['attributes'] and search in str(record['attributes'][field['field']]):
            decorator_instance = next((x for x in self.get_meta_decorators(data) if x['id'] == record['id']), None)
            if decorator_instance is None:
                self.get_meta_decorators(data).append({
                    'id': record['id'],
                    'search': [field['field']]
                })
            else:
                decorator_instance['search'].append(field['field'])

    def handle_search_decorator(self, data, Model, search):
        collection = Schema.get_collection(Model._meta.db_table)
        fields_to_search = self.get_fields_for_decorator_search(collection)

        for record in data['data']:
            for field in fields_to_search:
                self.handle_search_decorator_field(field, record, data, search)

    def get_meta_decorators(self, data):
        if 'meta' not in data or 'decorators' not in data['meta']:
            data['meta'] = {
                'decorators': []
            }
        return data['meta']['decorators']

    def decorators(self, data, Model, params):
        if 'search' in params and params['search']:
            self.handle_search_decorator(data, Model, params['search'])

        return data



from django.core.validators import RegexValidator
from django.db.models.fields import NOT_PROVIDED

VALIDATORS = {
    'max_length': {
        'type': 'is shorter than',
        'message': 'Ensure this value has at most %(limit_value)d characters'
    },
    'min_length': {
        'type': 'is longer than',
        'message': 'Ensure this value has at least %(limit_value)d characters'
    },
    'max_value': {
        'type': 'is less than',
        'message': 'Ensure this value is less than or equal to %(limit_value)s characters'
    },
    'min_value': {
        'type': 'is greater than',
        'message': 'Ensure this value is greater than or equal to %(limit_value)s characters'
    },
}


def add_validation(f, _type, message, value=None):
    validation = {
        'type': _type,
        'message': message,
    }
    if value is not None:
        validation['value'] = value
    if 'validations' not in f:
        f['validations'] = []
    f['validations'].append(validation)
    return f


def handle_validator(validator, f):
    if validator.code in VALIDATORS.keys():
        v = VALIDATORS[validator.code]
        message = v['message'] % {'limit_value': validator.limit_value}
        f = add_validation(f, v['type'], message, validator.limit_value)
    elif isinstance(validator, RegexValidator):
        message = validator.message
        if not isinstance(validator.message, str):
            message = 'Ensure this value match your pattern'
        f = add_validation(f, 'is like', message, validator.regex.pattern)
    return f


def handle_validators(validators, f):
    if len(validators):
        for validator in validators:
            if hasattr(validator, 'code'):
                f = handle_validator(validator, f)

    return f


def handle_is_present(field, f, blank, null):
    if (not blank or not null) and not (hasattr(field, 'default') and field.default != NOT_PROVIDED):
        f = add_validation(f, 'is present', 'Ensure this value is not null or not empty')

    return f

def _blank_flag(field):
    if hasattr(field, 'blank'):
        blank = field.blank
    else:
        #  some custom field doesn't set the blank flag. Django default value is false
        blank = False
    return blank

def _null_flag(field):
    if hasattr(field, 'null'):
        null = field.null
    else:
        #  some custom field doesn't set the null flag. Django default value is false
        null = False
    return null

def handle_validations(field, f):
    _type = None
    if hasattr(field, 'get_internal_type'):
        _type = field.get_internal_type()
    if not field.is_relation and not field.auto_created and _type != 'ArrayField':
        if hasattr(field, 'validators'):
            f = handle_validators(field.validators, f)
        f = handle_is_present(field, f, _blank_flag(field), _null_flag(field))

    return f



import copy
import sys
from unittest import mock

import pytest
from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema


# reset forest config dir auto import
from django_forest.utils.scope import ScopeManager


@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


@mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
@mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})    
@pytest.mark.usefixtures('reset_config_dir_import')
class ResourceCsvViewTests(TransactionTestCase):
    fixtures = ['question.json', 'choice.json', ]

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:csv', kwargs={'resource': 'tests_question'})
        self.reverse_url = reverse('django_forest:resources:csv', kwargs={'resource': 'tests_choice'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')
        ScopeManager.cache = {
            '1': {
                'scopes': {},
                'fetched_at': 'useless-here'
            }
        }

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    def test_get(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date,foo,bar',
            'fields[topic]': 'name',
            'search': '',
            'filters': '',
            'searchExtended': 0,
            'filename': 'questions',
            'header': 'id,topic,question text,pub date,foo,bar',
            'timezone': 'Europe/Paris'
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode('utf-8'),
                         'id,topic,question text,pub date,foo,bar\r\n1,,what is your favorite color?,2021-06-02T13:52:53.528000+00:00,what is your favorite color?+foo,what is your favorite color?+bar\r\n2,,do you like chocolate?,2021-06-02T15:52:53.528000+00:00,do you like chocolate?+foo,do you like chocolate?+bar\r\n3,,who is your favorite singer?,2021-06-03T13:52:53.528000+00:00,who is your favorite singer?+foo,who is your favorite singer?+bar\r\n')

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
    def test_get_related_data(self, *args, **kwargs):
        response = self.client.get(self.reverse_url, {
            'fields[tests_choice]': 'id,question,topic,choice_text',
            'fields[question]': 'question_text',
            'fields[topic]': 'name',
            'search': '',
            'filters': '',
            'searchExtended': 0,
            'filename': 'choices',
            'header': 'id,choice text',
            'timezone': 'Europe/Paris'
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode('utf-8'),
                         'id,choice text,,\r\n1,what is your favorite color?,,yes\r\n2,what is your favorite color?,,no\r\n3,do you like chocolate?,,good\r\n')

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.scope.ScopeManager._has_cache_expired', return_value=False)
    def test_wrong_operator(self, *args, **kwargs):
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,question_text,pub_date,foo,bar',
            'search': '',
            'filters': '{"field":"question_text","operator":"foo","value":"what is your favorite color?"}',
            'searchExtended': 0,
            'filename': 'questions',
            'header': 'id,question text,pub date,foo,bar',
            'timezone': 'Europe/Paris'
        })
        data = response.json()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(data, {
            'errors': [
                {
                    'detail': 'Unknown provided operator foo'
                }
            ]
        })



from django.test import TestCase
from django.urls import reverse


class AuthenticationLogoutViewTests(TestCase):
    def test_post(self):
        response = self.client.post(reverse('django_forest:authentication:logout'))
        self.assertEqual(response.status_code, 204)



from django.http import JsonResponse, HttpResponse

from django_forest.resources.utils.format import FormatFieldMixin
from django_forest.utils.schema import Schema
from django_forest.resources.utils.json_api_serializer import JsonApiSerializerMixin
from django_forest.resources.utils.resource import ResourceView
from django_forest.resources.utils.smart_field import SmartFieldMixin

from django_forest.utils.schema.json_api_schema import JsonApiSchema


class DetailView(SmartFieldMixin, FormatFieldMixin, JsonApiSerializerMixin, ResourceView):

    def get_instance(self, request, pk):
        # Notice: filter by scopes first
        queryset = self.Model.objects.all()
        scope_filters = self.get_scope(request, self.Model)
        if scope_filters is not None:
            queryset = queryset.filter(scope_filters)
        queryset = queryset.filter(pk=pk)

        if not len(queryset):
            raise Exception('Record does not exist or you don\'t have the right to query it')

        instance = queryset[0]
        # handle smart fields
        self.handle_smart_fields(instance, self.Model._meta.db_table)

        return instance

    def get(self, request, pk):
        try:
            instance = self.get_instance(request, pk)
        except Exception as e:
            return self.error_response(e)
        else:
            # handle smart fields
            self.handle_smart_fields(instance, self.Model._meta.db_table)

            # json api serializer
            include_data = self.get_include_data(Schema.get_collection(self.Model._meta.db_table)['fields'])
            JsonSchema = JsonApiSchema._registry[f'{self.Model._meta.db_table}Schema']
            data = JsonSchema(include_data=include_data).dump(instance)

            return JsonResponse(data, safe=False)

    def put(self, request, pk):
        body = self.get_body(request.body)

        try:
            attributes = self.populate_attribute(body)
            instance = self.get_instance(request, pk)
            for k, v in attributes.items():
                setattr(instance, k, v)
            instance = self.update_smart_fields(instance, body, self.Model._meta.db_table)
            instance.save()
        except Exception as e:
            return self.error_response(e)
        else:
            # Notice: one to one case, where a new object is created with a new pk
            # It needs to be deleted, as django orm will create a new object
            if str(instance.pk) != pk:
                self.Model.objects.filter(pk=pk).delete()

            # json api serializer
            Schema = JsonApiSchema._registry[f'{self.Model._meta.db_table}Schema']
            data = Schema().dump(instance)
            return JsonResponse(data, safe=False)

    def delete(self, request, pk):
        try:
            instance = self.get_instance(request, pk)
        except Exception as e:
            return self.error_response(e)
        else:
            instance.delete()
            return HttpResponse(status=204)



import logging
from urllib.parse import urljoin

from oic.oic import Client
from oic.oic.message import ProviderConfigurationResponse

from .configuration_retriever import retrieve
from .dynamic_client_registrator import register

logger = logging.getLogger(__name__)

class OidcClientManager:
    client = None

    @classmethod
    def get_client_for_callback_url(cls, callback_url):
        if cls.client:
            return cls.client

        configuration = retrieve()
        client_credentials = register({
            'token_endpoint_auth_method': 'none',
            'redirect_uris': [callback_url],
            'registration_endpoint': configuration['registration_endpoint']
        })
        logger.debug(f"client_credentials: {client_credentials}")
        client_data = {
            'client_id': client_credentials['client_id'],
            'issuer': configuration['issuer']
        }

        op_info = ProviderConfigurationResponse(
            client_id=client_data['client_id'],
            redirect_uri=callback_url,
            issuer=client_data['issuer'],
            authorization_endpoint=urljoin(client_data['issuer'], 'oidc/auth'),
            token_endpoint=urljoin(client_data['issuer'], 'oidc/token'),
            jwks_uri=urljoin(client_data['issuer'], 'oidc/jwks')
        )

        cls.client = Client(client_data['client_id'], verify_ssl=False)
        cls.client.handle_provider_config(op_info, op_info['issuer'])

        return cls.client



import copy
from marshmallow_jsonapi import fields

from django.test import TestCase

from django_forest.tests.fixtures.schema import test_schema
from django_forest.tests.models import Article, Session, Question, Choice

from django_forest.utils.collection import Collection
from django_forest.utils.schema.json_api_schema import JsonApiSchema, DjangoSchema
from django_forest.utils.schema import Schema
from django_forest.utils.scope import ScopeManager


class UtilsJsonApiSchemaTests(TestCase):
    fixtures = ['article.json', 'publication.json', 'session.json', 'question.json', 'choice.json']

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)

    def tearDown(self):
        # reset _registry after each test
        Collection._registry = {}
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    def test_json_api_schema_bad_name(self):
        Schema.handle_json_api_schema()
        with self.assertRaises(Exception) as cm:
            JsonApiSchema.get('Foo')
        self.assertEqual(cm.exception.args[0], 'The FooSchema does not exist in the JsonApiSchema. Make sure you correctly set it.')

    def test_json_api_schema(self):
        Schema.handle_json_api_schema()
        schema = JsonApiSchema.get('tests_question')
        question = Question.objects.get(pk=1)
        data = schema().dump(question)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_question',
                'id': 1,
                'relationships': {
                    'choice_set': {
                        'links': {
                            'related': '/forest/tests_question/1/relationships/choice_set'
                        }
                    },
                    'topic': {'links': {'related': '/forest/tests_question/1/relationships/topic'}}
                },
                'attributes': {
                    'pub_date': '2021-06-02T13:52:53.528000+00:00',
                    'question_text': 'what is your favorite color?'
                },
                'links': {
                    'self': '/forest/tests_question/1'
                }
            },
            'links': {
                'self': '/forest/tests_question/1'
            }
        })

    def test_json_api_schema_many(self):
        Schema.handle_json_api_schema()
        schema = JsonApiSchema.get('tests_question')
        queryset = Question.objects.all()
        data = schema().dump(queryset, many=True)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'relationships': {
                        'choice_set': {
                            'links': {
                                'related': '/forest/tests_question/1/relationships/choice_set'
                            }
                        },
                        'topic': {'links': {'related': '/forest/tests_question/1/relationships/topic'}}
                    },
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    }
                },
                {
                    'type': 'tests_question',
                    'id': 2,
                    'relationships': {
                        'choice_set': {
                            'links': {
                                'related': '/forest/tests_question/2/relationships/choice_set'
                            }
                        },
                        'topic': {'links': {'related': '/forest/tests_question/2/relationships/topic'}}
                    },
                    'attributes': {
                        'pub_date': '2021-06-02T15:52:53.528000+00:00',
                        'question_text': 'do you like chocolate?'
                    },
                    'links': {
                        'self': '/forest/tests_question/2'
                    }
                },
                {
                    'type': 'tests_question',
                    'id': 3,
                    'relationships': {
                        'choice_set': {
                            'links': {
                                'related': '/forest/tests_question/3/relationships/choice_set'
                            }
                        },
                        'topic': {'links': {'related': '/forest/tests_question/3/relationships/topic'}}
                    },
                    'attributes': {
                        'pub_date': '2021-06-03T13:52:53.528000+00:00',
                        'question_text': 'who is your favorite singer?'
                    },
                    'links': {
                        'self': '/forest/tests_question/3'
                    }
                }
            ]
        })

    def test_json_api_schema_foreign_key(self):
        Schema.handle_json_api_schema()
        schema = JsonApiSchema.get('tests_choice')
        choice = Choice.objects.get(pk=1)
        data = schema(include_data=['question']).dump(choice)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_choice',
                'attributes': {
                    'choice_text': 'yes',
                    'votes': 0
                },
                'id': 1,
                'relationships': {
                    'question': {
                        'links': {
                            'related': '/forest/tests_choice/1/relationships/question'
                        },
                        'data': {
                            'type': 'tests_question',
                            'id': '1'
                        }
                    }
                },
                'links': {
                    'self': '/forest/tests_choice/1'
                }
            },
            'links': {
                'self': '/forest/tests_choice/1'
            },
            'included': [
                {
                    'type': 'tests_question',
                    'id': 1,
                    'relationships': {
                        'choice_set': {
                            'links': {
                                'related': '/forest/tests_question/1/relationships/choice_set'
                            }
                        },
                        'topic': {'links': {'related': '/forest/tests_question/1/relationships/topic'}}
                    },
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?'
                    },
                    'links': {
                        'self': '/forest/tests_question/1'
                    }
                }
            ]
        })

    def test_json_api_schema_pk_is_not_id(self):
        Schema.handle_json_api_schema()
        schema = JsonApiSchema.get('tests_session')
        session = Session.objects.get(pk='foobar1234')
        data = schema().dump(session)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_session',
                'attributes': {
                    'session_key': 'foobar1234',
                    'expire_date': '2021-06-02T13:48:36.039000+00:00',
                    'session_data': 'foo',
                },
                'id': 'foobar1234',
                'links': {
                    'self': '/forest/tests_session/foobar1234'
                },
            },
            'links': {
                'self': '/forest/tests_session/foobar1234'
            },
        })

    def test_json_api_schema_many_to_many(self):
        Schema.handle_json_api_schema()
        schema = JsonApiSchema.get('tests_article')
        article = Article.objects.get(pk=1)
        data = schema(include_data=['publications']).dump(article)
        self.assertEqual(data, {
            'data': {
                'type': 'tests_article',
                'relationships': {
                    'publications': {
                        'links': {
                            'related': '/forest/tests_article/1/relationships/publications'
                        },
                        'data': [
                            {
                                'type': 'tests_publication',
                                'id': '1'
                            }
                        ]
                    }
                },
                'id': 1,
                'attributes': {
                    'headline': 'Django lets you build Web apps easily'
                },
                'links': {
                    'self': '/forest/tests_article/1'
                }
            },
            'links': {
                'self': '/forest/tests_article/1'
            },
            'included': [
                {
                    'type': 'tests_publication',
                    'relationships': {
                        'article_set': {
                            'links': {
                                'related': '/forest/tests_publication/1/relationships/article_set'
                            }
                        }
                    },
                    'id': 1,
                    'attributes': {
                        'title': 'The Python Journal'
                    },
                    'links': {
                        'self': '/forest/tests_publication/1'
                    }
                }
            ]
        })

    # These are dumb tests for having 100% covering
    # This is just some redefinition of Marshmallow Json Api for handling django
    def test_django_schema(self):
        with self.assertRaises(Exception) as cm:
            DjangoSchema()
        self.assertEqual(cm.exception.args[0], 'Must specify type_ class Meta option')

    def test_django_schema_no_url(self):
        with self.assertRaises(Exception) as cm:
            class FooSchema(DjangoSchema):
                class Meta:
                    type_ = 'foo'
                    self_url_kwargs = {'foo_id': f'<id>'}
            FooSchema()

        self.assertEqual(cm.exception.args[0], 'Must specify `self_url` Meta option when `self_url_kwargs` is specified')

    def test_django_schema_format_item_none(self):
        class FooSchema(DjangoSchema):
            class Meta:
                type_ = 'foo'

        schema = FooSchema()
        self.assertEqual(schema.format_item(None), None)



from enum import Enum
from typing import Any, Iterable


class Operator(Enum):
    lt = 'lt'
    lte = 'lte'
    gt = 'gt'
    gte = 'gte'


class BaseCondition:
    HAS_MANY = None

class Condition(BaseCondition):
    HAS_MANY = False

    def __init__(self, value: Any, *args, **kwargs):
        self.value = value

    def __repr__(self):
        return f"{self.value} {self.operator}"

    def __eq__(self, other) -> bool:
        if isinstance(self, other.__class__):
            return (
                self.value == other.value and self.operator == other.operator
            )
        return False

    @property
    def operator(self) -> Operator:
        raise NotImplementedError()


class Conditions(BaseCondition):
    HAS_MANY = True

    @property
    def conditions(self) -> Iterable:
        raise NotImplementedError()


class LowerThanCondition(Condition):
    @property
    def operator(self) -> Operator:
        return Operator.lt


class LowerThanOrEqualCondition(Condition):
    @property
    def operator(self) -> Operator:
        return Operator.lte


class GreaterThanCondition(Condition):
    @property
    def operator(self) -> Operator:
        return Operator.gt


class GreaterThanOrEqualCondition(Condition):
    @property
    def operator(self) -> Operator:
        return Operator.gte


class RangeCondition(Conditions):

    def __init__(self, start: Any, end: Any, *args, **kwargs):
        self.start = start
        self.end = end

    def __eq__(self, other) -> bool:
        if isinstance(self, other.__class__):
            return (
                self.start == other.start and self.end == other.end and self.conditions == other.conditions
            )
        return False

    @property
    def conditions(self) -> Iterable:
        return (
            GreaterThanOrEqualCondition(self.start),
            LowerThanCondition(self.end)
        )



# Generated by Django 3.2.4 on 2021-07-02 11:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0003_auto_20210701_1508'),
    ]

    operations = [
        migrations.CreateModel(
            name='Serial',
            fields=[
                ('uuid', models.UUIDField(primary_key=True, serialize=False)),
            ],
        ),
    ]



from django_forest.utils.views.base import BaseView
from .utils import get_row, execute_query
from django_forest.stats.utils.stats import StatsMixin

# TODO: support scopes once specification is achieved


class LiveQueriesView(StatsMixin, BaseView):
    def compute_data(self, query):
        data = {}
        for key, value in execute_query(query, 'key', 'value'):
            self.fill_data(data, key, int(value))
        return data

    def get_value(self, params, request, queryset=None):
        res = {}
        if params['type'] == 'Objective':
            value, objective = execute_query(params['query'], 'value', 'objective', one=True)
            res['objective'] = objective
            res['value'] = value
        else:
            res['countCurrent'] = get_row(params['query'])

        return res

    def get_pie(self, params, queryset=None):
        data = self.compute_data(params['query'])
        return [{
            'key': k,
            'value': v
        } for k, v in data.items()]

    def get_line(self, params, queryset=None):
        data = self.compute_data(params['query'])
        return [{
            'label': k,
            'values': {
                'value': v
            }
        } for k, v in data.items()]

    def get_leaderboard(self, params, queryset=None):
        data = self.compute_data(params['query'])
        return [{
            'key': k,
            'value': v
        } for k, v in data.items() if v is not None]

    def post(self, request, *args, **kwargs):
        params = self.get_body(request.body)
        return self.chart(params, request)



from datetime import datetime
try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

def get_timezone(string):
    return zoneinfo.ZoneInfo(string)

def get_utc_now():
    return get_now_aware_datetime(get_timezone('UTC'))

def get_now_aware_datetime(tz):
    return datetime.now(tz)



import copy
import json
import os
import sys

import django
import pytest
from unittest import mock
from django.test import TestCase, override_settings

from django_forest.tests.fixtures.schema import test_schema, test_choice_schema, \
    test_exclude_django_contrib_schema, test_serialized_schema, test_question_schema_data
from django_forest.tests.utils.test_forest_api_requester import mocked_requests, mocked_requests_no_data
from django_forest.utils.collection import Collection
from django_forest.utils.schema.json_api_schema import JsonApiSchema
from django_forest.utils.models import Models
from django_forest.utils.schema import Schema
from django_forest.utils.scope import ScopeManager

# reset forest config dir auto import
@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python3.8 or higher")
class UtilsSchemaTests(TestCase):

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)

    def tearDown(self):
        # reset _registry after each test
        Collection._registry = {}
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}
        Schema.schema_data = None
        Models.models = None

    @mock.patch.object(django, 'get_version', return_value='9.9.9')
    @mock.patch('importlib.metadata.version', return_value='0.0.0')
    def test_build_schema(self, mock_version, mock_orm_version):
        # reset schema
        Schema.schema = {
            'collections': [],
            'meta': {
                'liana': 'django-forestadmin',
                'liana_version': '0.0.0',
                'stack': {
                    'database_type': 'sqlite',
                    'orm_version': '9.9.9'
                },
            }
        }
        Schema.models = Models.list()
        schema = Schema.build_schema()
        self.assertEqual(schema, test_schema)

    @override_settings(FOREST={'INCLUDED_MODELS': ['tests_choice']})
    @mock.patch.object(django, 'get_version', return_value='9.9.9')
    @mock.patch('importlib.metadata.version', return_value='0.0.0')
    def test_build_schema_included_models(self, mock_version, mock_orm_version):
        # reset schema
        Schema.schema = {
            'collections': [],
            'meta': {
                'liana': 'django-forestadmin',
                'liana_version': '0.0.0',
                'stack': {
                    'database_type': 'sqlite',
                    'orm_version': '9.9.9'
                },
            }
        }
        Schema.models = Models.list(force=True)
        schema = Schema.build_schema()
        self.assertEqual(schema, test_choice_schema)

    @override_settings(FOREST={'EXCLUDED_MODELS': ['tests_permission', 'tests_group', 'tests_user', 'tests_contentType']})
    @mock.patch.object(django, 'get_version', return_value='9.9.9')
    @mock.patch('importlib.metadata.version', return_value='0.0.0')
    def test_build_schema_excluded_models(self, mock_version, mock_orm_version):
        # reset schema
        self.maxDiff = None
        Schema.schema = {
            'collections': [],
            'meta': {
                'liana': 'django-forestadmin',
                'liana_version': '0.0.0',
                'stack': {
                    'database_type': 'sqlite',
                    'orm_version': '9.9.9'
                },
            }
        }
        Schema.models = Models.list(force=True)
        schema = Schema.build_schema()
        self.assertEqual(schema, test_exclude_django_contrib_schema)

    @pytest.mark.usefixtures('reset_config_dir_import')
    @mock.patch('django_forest.utils.collection.Collection')
    def test_add_smart_features(self, collection_mock):
        Schema.add_smart_features()
        from django_forest.tests.forest import QuestionForest
        from django_forest.tests.models import Question
        collection_mock.register.assert_called_with(QuestionForest, Question)

    def test_get_collection(self):
        collection = Schema.get_collection('tests_question')
        self.assertEqual(collection, [x for x in test_schema['collections'] if x['name'] == 'tests_question'][0])

    def test_get_collection_inexist(self):
        collection = Schema.get_collection('Foo')
        self.assertEqual(collection, None)

    def test_handle_json_api_schema(self):
        Schema.handle_json_api_schema()
        self.assertEqual(len(JsonApiSchema._registry), 22)


# reset forest config dir auto import
@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


file_path = os.path.join(os.getcwd(), '.forestadmin-schema.json')


@pytest.fixture()
def dumb_forestadmin_schema():
    schema_data = json.dumps(test_schema, indent=2)
    with open(file_path, 'w') as f:
        f.write(schema_data)


@pytest.fixture()
def invalid_forestadmin_schema():
    schema_data = 'invalid'
    with open(file_path, 'w') as f:
        f.write(schema_data)


class UtilsSchemaFileTests(TestCase):
    def setUp(self):
        Schema.build_schema()
        Schema.add_smart_features()

    def tearDown(self):
        # reset _registry after each test
        Collection._registry = {}
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}
        Schema.schema_data = None
        Models.models = None
        if os.path.exists(file_path):
            os.remove(file_path)

    @pytest.mark.usefixtures('reset_config_dir_import')
    def test_handle_schema_file_no_file(self):
        with self.assertLogs() as cm:
            self.assertRaises(Exception, Schema.handle_schema_file())
            self.assertIsNone(Schema.schema_data)
            self.assertEqual(cm.output, [
                'ERROR:django_forest.utils.schema:The .forestadmin-schema.json file does not exist.',
                'ERROR:django_forest.utils.schema:The schema cannot be synchronized with Forest Admin servers.'
            ])

    @pytest.mark.usefixtures('reset_config_dir_import')
    @pytest.mark.usefixtures('dumb_forestadmin_schema')
    def test_handle_schema_file_production(self):
        Schema.handle_schema_file()
        self.assertIsNotNone(Schema.schema_data)

    @pytest.mark.usefixtures('reset_config_dir_import')
    @pytest.mark.usefixtures('invalid_forestadmin_schema')
    def test_handle_schema_file_invalid_json_production(self):
        with self.assertLogs() as cm:
            self.assertRaises(Exception, Schema.handle_schema_file())
            self.assertIsNone(Schema.schema_data)
            self.assertEqual(cm.output, [
                'ERROR:django_forest.utils.schema:The content of .forestadmin-schema.json file is not a correct JSON.',
                'ERROR:django_forest.utils.schema:The schema cannot be synchronized with Forest Admin servers.'
            ])

    @pytest.mark.usefixtures('reset_config_dir_import')
    @override_settings(DEBUG=True)
    def test_handle_schema_file_debug(self):
        Schema.handle_schema_file()
        with open(file_path, 'r') as f:
            data = f.read()
            data = json.loads(data)
            question = [c for c in data['collections'] if c['name'] == 'tests_question'][0]
            self.assertEqual(len(question['fields']), 7)
            foo_field = [f for f in question['fields'] if f['field'] == 'foo'][0]
            self.assertFalse('get' in foo_field)
            self.assertIsNotNone(Schema.schema_data)


class UtilsSchemaSendTests(TestCase):

    def test_get_serialized_schema(self):
        Schema.schema_data = test_question_schema_data
        serialized_schema = Schema.get_serialized_schema()
        self.assertEqual(serialized_schema, test_serialized_schema)

    @override_settings(FOREST={'FOREST_DISABLE_AUTO_SCHEMA_APPLY': True})
    @mock.patch.object(Schema, 'get_serialized_schema')
    def test_send_apimap_disable_apply(self, mocked_get_serialized_schema):
        Schema.send_apimap()
        mocked_get_serialized_schema.assert_not_called()

    @override_settings(FOREST={'FOREST_DISABLE_AUTO_SCHEMA_APPLY': 'foo'})
    def test_send_apimap_server_error(self):
        self.assertRaises(Exception, Schema.send_apimap())

    @override_settings(DEBUG=True)
    @mock.patch('requests.post', return_value=mocked_requests({'key1': 'value1'}, 200))
    def test_send_apimap(self, mocked_requests_post):
        Schema.schema_data = test_question_schema_data
        Schema.send_apimap()
        mocked_requests_post.assert_called_once_with(
            'https://api.test.forestadmin.com/forest/apimaps',
            data=json.dumps(test_serialized_schema),
            headers={'Content-Type': 'application/json', 'forest-secret-key': 'foo'},
            params={},
            verify=False
        )

    @override_settings(DEBUG=True)
    @mock.patch('requests.post', return_value=mocked_requests_no_data(204))
    def test_send_apimap_no_changes(self, mocked_requests_post):
        Schema.schema_data = test_question_schema_data
        Schema.send_apimap()
        mocked_requests_post.assert_called_once_with(
            'https://api.test.forestadmin.com/forest/apimaps',
            data=json.dumps(test_serialized_schema),
            headers={'Content-Type': 'application/json', 'forest-secret-key': 'foo'},
            params={},
            verify=False
        )

    @mock.patch('requests.post', return_value=mocked_requests({'key1': 'value1'}, 200))
    def test_send_apimap_production(self, mocked_requests_post):
        Schema.schema_data = test_question_schema_data
        Schema.send_apimap()
        mocked_requests_post.assert_called_once_with(
            'https://api.test.forestadmin.com/forest/apimaps',
            data=json.dumps(test_serialized_schema),
            headers={'Content-Type': 'application/json', 'forest-secret-key': 'foo'},
            params={},
        )

    @mock.patch('requests.post', return_value=mocked_requests({'warning': 'foo'}, 200))
    def test_send_apimap_warning(self, mocked_requests_post):
        Schema.schema_data = test_question_schema_data
        with self.assertLogs() as cm:
            Schema.send_apimap()
            self.assertEqual(cm.records[0].message, 'foo')
            self.assertEqual(cm.records[0].levelname, 'WARNING')

    @mock.patch('requests.post', side_effect=Exception('foo'))
    def test_send_apimap_zero(self, mocked_requests_post):
        Schema.schema_data = test_question_schema_data
        with self.assertLogs() as cm:
            self.assertRaises(Exception, Schema.send_apimap())
            self.assertEqual(cm.records[0].message,
                             'Cannot send the apimap to Forest. Are you online?')
            self.assertEqual(cm.records[0].levelname, 'WARNING')

    @mock.patch('requests.post', return_value=mocked_requests({}, 404))
    def test_send_apimap_not_found(self, mocked_requests_post):
        Schema.schema_data = test_question_schema_data
        with self.assertLogs() as cm:
            Schema.send_apimap()
            self.assertEqual(cm.records[0].message,
                             'Cannot find the project related to the envSecret you configured. Can you check on Forest that you copied it properly in the Forest settings?')
            self.assertEqual(cm.records[0].levelname, 'ERROR')

    @mock.patch('requests.post', return_value=mocked_requests({}, 503))
    def test_send_apimap_unavailable(self, mocked_requests_post):
        Schema.schema_data = test_question_schema_data
        with self.assertLogs() as cm:
            Schema.send_apimap()
            self.assertEqual(cm.records[0].message,
                             'Forest is in maintenance for a few minutes. We are upgrading your experience in the forest. We just need a few more minutes to get it right.')
            self.assertEqual(cm.records[0].levelname, 'WARNING')

    @mock.patch('requests.post', return_value=mocked_requests({}, 500))
    def test_send_apimap_error(self, mocked_requests_post):
        Schema.schema_data = test_question_schema_data

        with self.assertLogs() as cm:
            Schema.send_apimap()
            self.assertEqual(cm.records[0].message,
                             'An error occured with the apimap sent to Forest. Please contact support@forestadmin.com for further investigations.')
            self.assertEqual(cm.records[0].levelname, 'ERROR')


class UtilsSchemaInitTests(TestCase):
    def test_schema_meta(self):
        self.assertTrue('liana' in Schema.schema['meta'])
        self.assertTrue('liana_version' in Schema.schema['meta'])
        self.assertTrue('stack' in Schema.schema['meta'])
        self.assertTrue('database_type' in Schema.schema['meta']['stack'])
        self.assertTrue('orm_version' in Schema.schema['meta']['stack'])


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python3.8 or higher")
class UtilsGetAppVersionTests(TestCase):
    @mock.patch('importlib.metadata.version', return_value='0.0.1')
    def test_get_app_version(self, mock_version):
        from django_forest.utils.schema.version import get_app_version
        version = get_app_version()
        self.assertEqual(version, '0.0.1')

    @mock.patch('importlib.metadata.version', side_effect=Exception('error'))
    def test_get_app_version_error(self, mock_version):
        from django_forest.utils.schema.version import get_app_version
        version = get_app_version()
        self.assertEqual(version, '0.0.0')


@pytest.mark.skipif(sys.version_info >= (3, 8), reason="requires python3.7 or lower")
class UtilsGetAppOldPythonTests(TestCase):

    @mock.patch('importlib_metadata.version', return_value='0.0.1')
    def test_get_app_version(self, mock_version):
        from django_forest.utils.schema.version import get_app_version
        version = get_app_version()
        self.assertEqual(version, '0.0.1')

    @mock.patch('importlib_metadata.version', side_effect=Exception('error'))
    def test_get_app_version_error(self, mock_version):
        from django_forest.utils.schema.version import get_app_version
        version = get_app_version()
        self.assertEqual(version, '0.0.0')



import os
import urllib3
from distutils.util import strtobool

from django.apps import AppConfig

disable_warnings = os.getenv('URLLIB3_DISABLE_WARNINGS', 'False')
try:
    disable_warnings = strtobool(disable_warnings)
except Exception:
    disable_warnings = False
finally:
    if disable_warnings:
        urllib3.disable_warnings()


class BaseForestException(Exception):
    pass


class ForestConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_forest'



import json
import logging

from oic.oic.message import AuthorizationRequest

from django.http import JsonResponse
from django.views.generic import View
from django_forest.authentication.exception import AuthenticationClientException

from django_forest.authentication.oidc.client_manager import OidcClientManager
from django_forest.authentication.utils import authentication_exception, get_callback_url
from django_forest.utils.error_handler import MESSAGES


logger = logging.getLogger(__name__)

# Based on https://pyoidc.readthedocs.io/en/latest/examples/rp.html


class IndexView(View):

    def _get_payload(self, request):
        try:
            payload = json.loads(request.body)
        except json.JSONDecodeError:
            raise AuthenticationClientException(
                'request.body is not a valid json object'
            )
        return payload

    def _get_rendering_id(self, payload):
        try:
            rendering_id = payload['renderingId']
        except KeyError:
            raise AuthenticationClientException(
                MESSAGES['SERVER_TRANSACTION']['MISSING_RENDERING_ID']
            )
        if not any(isinstance(rendering_id, _type) for _type in [str, int]):
            raise AuthenticationClientException(MESSAGES['SERVER_TRANSACTION']['INVALID_RENDERING_ID'])
        return int(rendering_id)

    def _get_authorization_url(self, redirect_url, state):
        client = OidcClientManager.get_client_for_callback_url(redirect_url)
        logger.debug(f"client_id: {client.client_id}")
        args = {
            'client_id': client.client_id,
            'response_type': 'code',
            'scope': ['openid', 'email', 'profile'],
            'state': json.dumps(state),
            'redirect_uri': redirect_url,
        }
        auth_req: AuthorizationRequest = client.construct_AuthorizationRequest(request_args=args)
        authorization_url = auth_req.request(client.authorization_endpoint)

        return authorization_url

    @authentication_exception
    def post(self, request, *args, **kwargs):
        payload = self._get_payload(request)
        rendering_id = self._get_rendering_id(payload)
        callback_url = get_callback_url()
        authorization_url = self._get_authorization_url(callback_url, {'renderingId': rendering_id})
        return JsonResponse({'authorizationUrl': authorization_url})



import sys
from distutils.util import strtobool
from uuid import UUID

from django.db.models import Q

from django_forest.utils.collection import Collection
from ..in_search_fields import in_search_fields
from django_forest.utils.schema import Schema


class SearchMixin:
    def get_fields_to_search(self, collection):
        fields_to_search = []
        for x in collection['fields']:
            if x['type'] in ('String', 'Number', 'Enum') \
                    and not x['reference'] \
                    and not x['is_virtual'] \
                    and in_search_fields(x['field'], collection['search_fields']):
                fields_to_search.append(x)
        return fields_to_search

    def is_number(self, search):
        is_number = True
        try:
            search = int(search)
        except ValueError:
            try:
                search = float(search)
            except ValueError:
                is_number = False

        return search, is_number

    def handle_number(self, search, lookup_field):
        q_object = Q()

        search, is_number = self.is_number(search)

        # Notice, only add condition if value is number
        if is_number:
            # Notice: use LIKE operator when too big
            if search <= sys.maxsize:
                q_object = Q(**{lookup_field: search})
            else:
                q_object = Q(**{f'{lookup_field}__contains': search})

        return q_object

    def is_uuid(self, search):
        try:
            UUID(search)
        except ValueError:
            return False
        else:
            return True

    def handle_string(self, search, lookup_field):
        is_uuid = self.is_uuid(search)

        if is_uuid:
            q_object = Q(**{lookup_field: search})
        else:
            q_object = Q(**{f'{lookup_field}__contains': search})

        return q_object

    def handle_enum(self, search, enums, lookup_field):
        q_object = Q()

        # Notice: only add condition if search in enums
        if search in enums:
            q_object = Q(**{lookup_field: search})

        return q_object

    def get_lookup_field(self, field_name, related_field):
        lookup_field = field_name
        if related_field is not None:
            lookup_field = f'{related_field}__{lookup_field}'
        return lookup_field

    def handle_field(self, search, field, related_field_name=None):
        q_objects = Q()

        lookup_field = self.get_lookup_field(field['field'], related_field_name)
        if field['type'] == 'Enum':
            q_objects |= self.handle_enum(search, field['enums'], lookup_field)
        elif field['type'] == 'Number':
            q_objects |= self.handle_number(search, lookup_field)
        else:
            q_objects |= self.handle_string(search, lookup_field)

        return q_objects

    def handle_search_extended(self, search, Model):
        q_objects = Q()

        related_fields = [x for x in Model._meta.get_fields() if x.is_relation and not x.many_to_many]
        for related_field in related_fields:
            q_objects |= self.fill_conditions(search, related_field.related_model._meta.db_table, related_field.name)

        return q_objects

    def add_smart_field(self, smart_field, resource, search):
        q_object = Q()
        method = smart_field['search']
        if isinstance(method, str):
            q_object = getattr(Collection._registry[resource], method)(search)
        elif callable(method):
            q_object = method(search)
        return q_object

    def add_smart_fields(self, collection, resource, search):
        q_objects = Q()
        smart_fields = [x for x in collection['fields'] if x['is_virtual']]
        for smart_field in smart_fields:
            if 'search' in smart_field:
                q_objects |= self.add_smart_field(smart_field, resource, search)

        return q_objects

    def fill_conditions(self, search, resource, related_field_name=None):
        q_objects = Q()

        collection = Schema.get_collection(resource)
        fields_to_search = self.get_fields_to_search(collection)
        for field in fields_to_search:
            q_objects |= self.handle_field(search, field, related_field_name)

        # Notice handle smart fields
        q_objects |= self.add_smart_fields(collection, resource, search)

        return q_objects

    def get_search(self, params, Model):
        q_objects = Q()
        search = params['search']

        q_objects |= self.fill_conditions(search, Model._meta.db_table)

        if 'searchExtended' in params and strtobool(str(params['searchExtended'])):
            q_objects |= self.handle_search_extended(search, Model)

        return q_objects



from django.http import JsonResponse

from django_forest.utils.views.base import BaseView


class ResourceView(BaseView):
    def dispatch(self, request, resource, *args, **kwargs):
        try:
            self.Model = self.get_model(resource)
        except Exception as e:
            return self.error_response(e)
        else:
            return super().dispatch(request, *args, **kwargs)

    def get_count(self, queryset, params, request):
        try:
            # enhance queryset (compute scope)
            queryset = self.enhance_queryset(queryset, self.Model, params, request)
        except Exception as e:
            return self.error_response(e)
        else:
            return JsonResponse({'count': queryset.count()}, safe=False)



from django.test import TestCase
from django.urls import reverse


class IndexViewTests(TestCase):
    def test_get(self):
        response = self.client.get(reverse('django_forest:index'))
        self.assertEqual(response.status_code, 204)



from django_forest.apps import BaseForestException


class BaseAuthenticationException(BaseForestException):
    pass

class AuthenticationClientException(BaseAuthenticationException):
    STATUS = 401


class AuthenticationSettingsException(BaseAuthenticationException):
    STATUS = 500


class AuthenticationThirdPartyException(BaseAuthenticationException):
    STATUS = 503



class LimitFieldsMixin:
    def handle_fields(self, params, lookup, Model, queryset):
        args = []
        fields_name = [x.name for x in Model._meta.get_fields()]
        for param in params[lookup].split(','):
            if param in fields_name:
                args.append(param)

        return queryset.only(*args)

    def handle_context(self, Model, queryset):
        args = []
        for field in Model._meta.get_fields():
            if not field.is_relation or (field.many_to_one or field.one_to_one):
                args.append(field.name)

        return queryset.only(*args)

    def handle_limit_fields(self, params, Model, queryset):

        lookup = f'fields[{Model._meta.db_table}]'
        if lookup in params:
            queryset = self.handle_fields(params, lookup, Model, queryset)
        return queryset
